from django.contrib.gis.geos import MultiPolygon, Polygon
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction

from codelist.models import KonovaCode
from codelist.settings import CODELIST_CONSERVATION_OFFICE_ID, CODELIST_COMPENSATION_HANDLER_ID
from compensation.models import EcoAccount, EcoAccountDocument
from intervention.models import Responsibility, Handler
from konova.management.commands.kspMigrater.compensation_migrater import CompensationMigrater
from konova.models import Geometry
from konova.sub_settings.lanis_settings import DEFAULT_SRID_RLP, DEFAULT_SRID


class EcoAccountMigrater(CompensationMigrater):

    def migrate(self):
        cursor = self.db_connection.cursor()
        cursor.execute(
            'select '
                'om."KENNUNG", '
                'linf."OBJBEZ", '
                'ST_AsEWKT(ST_Multi(ST_CollectionExtract(ST_MakeValid(ST_Transform(geomf.the_geom,4326)), 3))) as geomf, '
                'ST_AsEWKT(ST_Multi(ST_CollectionExtract(ST_MakeValid(ST_Transform(geoml.the_geom,4326)), 2))) as geoml, '
                'ST_AsEWKT(ST_Multi(ST_CollectionExtract(ST_MakeValid(ST_Transform(geomp.the_geom,4326)), 1))) as geomp, '
                'linf."Bemerkung" '
            'from "OBJ_MASTER" om '
                'left join "LINFOS" linf on om."GISPADID"=linf."GISPADID" '
                'left join kom on om."GISPADID"=kom.gispadid '
                'left join geometry_f geomf on om."GISPADID"=geomf.gispadid '
                'left join geometry_l geoml on om."GISPADID"=geoml.gispadid '
                'left join geometry_p geomp on om."GISPADID"=geomp.gispadid '
                'left join "Aufwertung" auf on om."GISPADID"=auf."GISPADID" '
            'where '
                'om."OKL"=7730081 and '
                'om.archiv=false and '
                'om.nicht_vollstaendig=0 '
        )

        all_oeks = cursor.fetchall()
        len_all_oeks = len(all_oeks)
        num_processed = 0
        print(f"Migrate OEKs to ecoaccounts...")
        print(f"--Found {len_all_oeks} entries. Process now...")
        for oek in all_oeks:
            if num_processed % 500 == 0:
                print(f"----{num_processed}/{len_all_oeks} processed")

            with transaction.atomic():
                oek_identifier = oek[0]
                oek_title = oek[1]
                oek_comment = oek[5]
                eco_account = EcoAccount.objects.get_or_create(
                    identifier=oek_identifier
                )[0]

                eco_account.title = oek_title
                eco_account.comment = oek_comment
                eco_account = self._migrate_geometry(eco_account, oek)
                eco_account = self._migrate_responsibility(eco_account, oek)
                eco_account = self._migrate_states(eco_account, oek)
                eco_account = self._migrate_deadlines(eco_account, oek)
                eco_account = self._migrate_action_control_deadlines(eco_account, oek)
                eco_account = self._migrate_actions(eco_account, oek)
                eco_account = self._migrate_documents(eco_account, EcoAccountDocument, oek)
                eco_account.save()

                num_processed += 1
        cursor.close()

    def _migrate_geometry(self, instance, db_result: tuple):
        identifier = f"'{db_result[0]}'"
        empty_str = f"''"
        tmp_cursor = self.db_connection.cursor()
        tmp_cursor.execute(
            'select '
                'ST_AsEWKT(ST_Multi(ST_CollectionExtract(ST_MakeValid(ST_Union(ST_Transform(geom.the_geom,4326)))))) '
            'from "OBJ_MASTER" om '
                'left join "Aufwertung" auf on om."GISPADID"=auf."GISPADID" '
                'left join "OBJ_MASTER" om_kom on auf."Infos"=om_kom."KENNUNG" '
                'left join geometry_f geom on geom.gispadid=om_kom."GISPADID" '
            'where '
                'om."OKL"=7730081 and '
                'om.archiv=false and '
                'om_kom.archiv=false and '
                f'(auf."Infos" is not null and auf."Infos"!={empty_str}) and '
                f'om."KENNUNG"={identifier}'
        )
        deduction_db_results = tmp_cursor.fetchall()
        if len(deduction_db_results) != 1:
            raise AssertionError(f"Unexpected number of matches: {deduction_db_results}")
        db_result_geom = deduction_db_results[0][0]
        if db_result_geom is not None:
            deductions_db_result_geom = MultiPolygon.from_ewkt(deduction_db_results[0][0])
        else:
            deductions_db_result_geom = MultiPolygon(srid=DEFAULT_SRID)

        tmp_cursor.execute(
            'select '
                'ST_AsEWKT(ST_Multi(ST_CollectionExtract(ST_MakeValid(ST_Transform(geomf.the_geom,4326)), 3))) as geomf '
            'from "OBJ_MASTER" om '
                'left join geometry_f geomf on om."GISPADID"=geomf.gispadid '
            'where '
                f'om."KENNUNG"={identifier}'
        )
        original_geom_db_results = tmp_cursor.fetchall()
        if len(original_geom_db_results) != 1:
            raise AssertionError(f"Unexpected number of matches: {original_geom_db_results}")
        account_db_result_geom = MultiPolygon.from_ewkt(original_geom_db_results[0][0])

        db_result_geom = account_db_result_geom.union(deductions_db_result_geom)
        if isinstance(db_result_geom, Polygon):
            db_result_geom = MultiPolygon(db_result_geom, srid=DEFAULT_SRID)
        instance.geometry = instance.geometry or Geometry()
        try:
            # Calculate area by transforming
            rlp_geom = db_result_geom.transform(ct=DEFAULT_SRID_RLP, clone=True)
            area = int(rlp_geom.area)
            instance.deductable_surface = area
            instance.geometry.geom = db_result_geom if not db_result_geom.empty else None
            instance.geometry.save()
        except TypeError:
            raise TypeError(f"{identifier}, {db_result_geom}")
        tmp_cursor.close()
        return instance

    def _migrate_responsibility(self, eco_account, oek):
        acc_identifier = f"'{oek[0]}'"
        tmp_cursor = self.db_connection.cursor()
        tmp_cursor.execute(
            'select '
                'adr."adr_pruef" as ets, '
                'linf."AZ", '
                'oek.traeger, '
                'oek.bemerkungtraeger '
            'from "OBJ_MASTER" om '
                'left join "LINFOS" linf on om."GISPADID"=linf."GISPADID" '
                'left join adressrolle adr on adr."GISPADID"=om."GISPADID" '
                'left join oek on om."GISPADID"=oek.gispadid '
            'where '
                f'om."KENNUNG"={acc_identifier}'
        )
        db_results = tmp_cursor.fetchall()
        if len(db_results) != 1:
            raise AssertionError(f"{acc_identifier} has invalid responsibilities: {db_results}")

        db_results = db_results[0]
        cons_office_code = db_results[0]
        cons_file_number = db_results[1]
        handler_type = db_results[2]
        handler_detail = db_results[3]

        eco_account.responsible = eco_account.responsible or Responsibility.objects.create()
        try:
            conservation_office = KonovaCode.objects.get(
                atom_id=cons_office_code,
                is_selectable=True,
                is_archived=False,
                code_lists__in=[CODELIST_CONSERVATION_OFFICE_ID]
            )
        except ObjectDoesNotExist:
            raise ObjectDoesNotExist(f"{acc_identifier}, {db_results}")
        try:
            handler_type = KonovaCode.objects.get(
                atom_id=handler_type,
                is_selectable=True,
                is_archived=False,
                code_lists__in=[CODELIST_COMPENSATION_HANDLER_ID]
            )
        except ObjectDoesNotExist:
            handler_type = None

        eco_account.responsible.conservation_office = conservation_office
        eco_account.responsible.conservation_file_number = cons_file_number
        handler = eco_account.responsible.handler or Handler.objects.create()
        handler.type = handler_type
        handler.detail = handler_detail
        eco_account.responsible.handler = handler
        eco_account.responsible.handler.save()
        eco_account.responsible.save()

        tmp_cursor.close()
        return eco_account